@@ -26,9 +26,10 @@ import os.path
 import sys
 import yaml
 import time
+import re
+import shutil
 
-from collections import defaultdict
-from jinja2 import Environment
+from jinja2 import Environment, FileSystemLoader
 
 import ansible.constants as C
 from ansible.cli import CLI
@@ -39,7 +40,7 @@ from ansible.galaxy.role import GalaxyRole
 from ansible.galaxy.login import GalaxyLogin
 from ansible.galaxy.token import GalaxyToken
 from ansible.playbook.role.requirement import RoleRequirement
-from ansible.module_utils._text import to_bytes, to_text
+from ansible.module_utils._text import to_text
 
 try:
     from __main__ import display
@@ -79,12 +80,16 @@ class GalaxyCLI(CLI):
             self.parser.set_usage("usage: %prog import [options] github_user github_repo")
             self.parser.add_option('--no-wait', dest='wait', action='store_false', default=True, help='Don\'t wait for import results.')
             self.parser.add_option('--branch', dest='reference', help='The name of a branch to import. Defaults to the repository\'s default branch (usually master)')
+            self.parser.add_option('--role-name', dest='role_name', help='The name the role should have, if different than the repo name')
             self.parser.add_option('--status', dest='check_status', action='store_true', default=False, help='Check the status of the most recent import request for given github_user/github_repo.')
         elif self.action == "info":
             self.parser.set_usage("usage: %prog info [options] role_name[,version]")
         elif self.action == "init":
             self.parser.set_usage("usage: %prog init [options] role_name")
             self.parser.add_option('-p', '--init-path', dest='init_path', default="./", help='The path in which the skeleton role will be created. The default is the current working directory.')
+            self.parser.add_option('--container-enabled', dest='container_enabled', action='store_true', default=False,
+                                   help='Initialize the skeleton role with default contents for a Container Enabled role.')
+            self.parser.add_option('--role-skeleton', dest='role_skeleton', default=None, help='The path to a role skeleton that the new role should be based upon.')
         elif self.action == "install":
             self.parser.set_usage("usage: %prog install [options] [-r FILE | role_name(s)[,version] | scm+role_repo_url[,version] | tar_file(s)]")
             self.parser.add_option('-i', '--ignore-errors', dest='ignore_errors', action='store_true', default=False, help='Ignore errors and continue with the next specified role.')
@@ -100,7 +105,7 @@ class GalaxyCLI(CLI):
         elif self.action == "search":
             self.parser.set_usage("usage: %prog search [searchterm1 searchterm2] [--galaxy-tags galaxy_tag1,galaxy_tag2] [--platforms platform1,platform2] [--author username]")
             self.parser.add_option('--platforms', dest='platforms', help='list of OS platforms to filter by')
-            self.parser.add_option('--galaxy-tags', dest='tags', help='list of galaxy tags to filter by')
+            self.parser.add_option('--galaxy-tags', dest='galaxy_tags', help='list of galaxy tags to filter by')
             self.parser.add_option('--author', dest='author', help='GitHub username')
         elif self.action == "setup":
             self.parser.set_usage("usage: %prog setup [options] source github_user github_repo secret")
@@ -120,10 +125,10 @@ class GalaxyCLI(CLI):
         if self.action in ("init","install"):
             self.parser.add_option('-f', '--force', dest='force', action='store_true', default=False, help='Force overwriting an existing role')
 
-        self.options, self.args = self.parser.parse_args(self.args[1:])
+        super(GalaxyCLI, self).parse()
+
         display.verbosity = self.options.verbosity
         self.galaxy = Galaxy(self.options)
-        return True
 
     def run(self):
 
@@ -173,7 +178,7 @@ class GalaxyCLI(CLI):
 
         init_path  = self.get_opt('init_path', './')
         force      = self.get_opt('force', False)
-        offline    = self.get_opt('offline', False)
+        role_skeleton = self.get_opt('role_skeleton', C.GALAXY_ROLE_SKELETON)
 
         role_name = self.args.pop(0).strip() if self.args else None
         if not role_name:
@@ -188,79 +193,54 @@ class GalaxyCLI(CLI):
                             "however it will reset any main.yml files that may have\n"
                             "been modified there already." % role_path)
 
-        # create default README.md
+        inject_data = dict(
+            role_name=role_name,
+            author='your name',
+            description='your description',
+            company='your company (optional)',
+            license='license (GPLv2, CC-BY, etc)',
+            issue_tracker_url='http://example.com/issue/tracker',
+            min_ansible_version='1.2',
+            container_enabled=self.options.container_enabled
+        )
+
+        # create role directory
         if not os.path.exists(role_path):
             os.makedirs(role_path)
-        readme_path = os.path.join(role_path, "README.md")
-        f = open(readme_path, "wb")
-        f.write(to_bytes(self.galaxy.default_readme))
-        f.close()
 
-        # create default .travis.yml
-        travis = Environment().from_string(self.galaxy.default_travis).render()
-        f = open(os.path.join(role_path, '.travis.yml'), 'w')
-        f.write(travis)
-        f.close()
+        if role_skeleton is not None:
+            skeleton_ignore_expressions = C.GALAXY_ROLE_SKELETON_IGNORE
+        else:
+            role_skeleton = self.galaxy.default_role_skeleton_path
+            skeleton_ignore_expressions = ['^.*/.git_keep$']
 
-        for dir in GalaxyRole.ROLE_DIRS:
-            dir_path = os.path.join(init_path, role_name, dir)
-            main_yml_path = os.path.join(dir_path, 'main.yml')
+        role_skeleton = os.path.expanduser(role_skeleton)
+        skeleton_ignore_re = list(map(lambda x: re.compile(x), skeleton_ignore_expressions))
 
-            # create the directory if it doesn't exist already
-            if not os.path.exists(dir_path):
-                os.makedirs(dir_path)
+        template_env = Environment(loader=FileSystemLoader(role_skeleton))
 
-            # now create the main.yml file for that directory
-            if dir == "meta":
-                # create a skeleton meta/main.yml with a valid galaxy_info
-                # datastructure in place, plus with all of the available
-                # platforms included (but commented out), the galaxy_tags
-                # list, and the dependencies section
-                platforms = []
-                if not offline:
-                    platforms = self.api.get_list("platforms") or []
+        for root, dirs, files in os.walk(role_skeleton, topdown=True):
+            rel_root = os.path.relpath(root, role_skeleton)
+            in_templates_dir = rel_root.split(os.sep, 1)[0] == 'templates'
+            dirs[:] = filter(lambda d: not any(map(lambda r: r.match(os.path.join(rel_root, d)), skeleton_ignore_re)), dirs)
 
-                # group the list of platforms from the api based
-                # on their names, with the release field being
-                # appended to a list of versions
-                platform_groups = defaultdict(list)
-                for platform in platforms:
-                    platform_groups[platform['name']].append(platform['release'])
-                    platform_groups[platform['name']].sort()
+            for f in files:
+                filename, ext = os.path.splitext(f)
+                if any(map(lambda r: r.match(os.path.join(rel_root, f)), skeleton_ignore_re)):
+                    continue
+                elif ext == ".j2" and not in_templates_dir:
+                    src_template = os.path.join(rel_root, f)
+                    dest_file = os.path.join(role_path, rel_root, filename)
+                    template_env.get_template(src_template).stream(inject_data).dump(dest_file)
+                else:
+                    f_rel_path = os.path.relpath(os.path.join(root, f), role_skeleton)
+                    shutil.copyfile(os.path.join(root, f), os.path.join(role_path, f_rel_path))
 
-                inject = dict(
-                    author = 'your name',
-                    description = 'your description',
-                    company = 'your company (optional)',
-                    license = 'license (GPLv2, CC-BY, etc)',
-                    issue_tracker_url = 'http://example.com/issue/tracker',
-                    min_ansible_version = '1.2',
-                    platforms = platform_groups,
-                )
-                rendered_meta = Environment().from_string(self.galaxy.default_meta).render(inject)
-                f = open(main_yml_path, 'w')
-                f.write(rendered_meta)
-                f.close()
-                pass
-            elif dir == "tests":
-                # create tests/test.yml
-                inject = dict(
-                    role_name = role_name
-                )
-                playbook = Environment().from_string(self.galaxy.default_test).render(inject)
-                f = open(os.path.join(dir_path, 'test.yml'), 'w')
-                f.write(playbook)
-                f.close()
+            for d in dirs:
+                dir_path = os.path.join(role_path, rel_root, d)
+                if not os.path.exists(dir_path):
+                    os.makedirs(dir_path)
 
-                # create tests/inventory
-                f = open(os.path.join(dir_path, 'inventory'), 'w')
-                f.write('localhost')
-                f.close()
-            elif dir not in ('files','templates'):
-                # just write a (mostly) empty YAML file for main.yml
-                f = open(main_yml_path, 'w')
-                f.write('---\n# %s file for %s\n' % (dir,role_name))
-                f.close()
         display.display("- %s was created successfully" % role_name)
 
     def execute_info(self):
@@ -376,7 +356,7 @@ class GalaxyCLI(CLI):
                         roles_left.append(GalaxyRole(self.galaxy, **role))
                 f.close()
             except (IOError, OSError) as e:
-                display.error('Unable to open %s: %s' % (role_file, str(e)))
+                raise AnsibleError('Unable to open %s: %s' % (role_file, str(e)))
         else:
             # roles were specified directly, so we'll just go out grab them
             # (and their dependencies, unless the user doesn't want us to).
@@ -388,9 +368,20 @@ class GalaxyCLI(CLI):
             display.vvv('Installing role %s ' % role.name)
             # query the galaxy API for the role data
 
-            if role.install_info is not None and not force:
-                display.display('- %s is already installed, skipping.' % role.name)
-                continue
+            if role.install_info is not None:
+                if role.install_info['version'] != role.version:
+                    if force:
+                        display.display('- changing role %s from %s to %s' %
+                                        (role.name, role.install_info['version'], role.version or "unspecified"))
+                        role.remove()
+                    else:
+                        display.warning('- %s (%s) is already installed - use --force to change version to %s' %
+                                        (role.name, role.install_info['version'], role.version or "unspecified"))
+                        continue
+                else:
+                    if not force:
+                        display.display('- %s is already installed, skipping.' % str(role))
+                        continue
 
             try:
                 installed = role.install()
@@ -411,14 +402,18 @@ class GalaxyCLI(CLI):
                         # we know we can skip this, as it's not going to
                         # be found on galaxy.ansible.com
                         continue
-                    if dep_role.install_info is None or force:
+                    if dep_role.install_info is None:
                         if dep_role not in roles_left:
-                            display.display('- adding dependency: %s' % dep_role.name)
+                            display.display('- adding dependency: %s' % str(dep_role))
                             roles_left.append(dep_role)
                         else:
                             display.display('- dependency %s already pending installation.' % dep_role.name)
                     else:
-                        display.display('- dependency %s is already installed, skipping.' % dep_role.name)
+                        if dep_role.install_info['version'] != dep_role.version:
+                            display.warning('- dependency %s from role %s differs from already installed version (%s), skipping' %
+                                            (str(dep_role), role.name, dep_role.install_info['version']))
+                        else:
+                            display.display('- dependency %s is already installed, skipping.' % dep_role.name)
 
             if not installed:
                 display.warning("- %s was NOT installed successfully." % role.name)
@@ -505,11 +500,11 @@ class GalaxyCLI(CLI):
                 terms.append(self.args.pop())
             search = '+'.join(terms[::-1])
 
-        if not search and not self.options.platforms and not self.options.tags and not self.options.author:
+        if not search and not self.options.platforms and not self.options.galaxy_tags and not self.options.author:
             raise AnsibleError("Invalid query. At least one search term, platform, galaxy tag or author must be provided.")
 
         response = self.api.search_roles(search, platforms=self.options.platforms,
-            tags=self.options.tags, author=self.options.author, page_size=page_size)
+            tags=self.options.galaxy_tags, author=self.options.author, page_size=page_size)
 
         if response['count'] == 0:
             display.display("No roles match your search.", color=C.COLOR_ERROR)
@@ -540,7 +535,7 @@ class GalaxyCLI(CLI):
 
     def execute_login(self):
         """
-        Verify user's identify via Github and retreive an auth token from Galaxy.
+        Verify user's identify via Github and retrieve an auth token from Galaxy.
         """
         # Authenticate with github and retrieve a token
         if self.options.token is None:
@@ -559,7 +554,7 @@ class GalaxyCLI(CLI):
         token = GalaxyToken()
         token.set(galaxy_response['token'])
 
-        display.display("Succesfully logged into Galaxy as %s" % galaxy_response['username'])
+        display.display("Successfully logged into Galaxy as %s" % galaxy_response['username'])
         return 0
 
     def execute_import(self):
@@ -585,7 +580,7 @@ class GalaxyCLI(CLI):
             task = self.api.get_import_task(github_user=github_user, github_repo=github_repo)
         else:
             # Submit an import request
-            task = self.api.create_import_task(github_user, github_repo, reference=self.options.reference)
+            task = self.api.create_import_task(github_user, github_repo, reference=self.options.reference, role_name=self.options.role_name)
 
             if len(task) > 1:
                 # found multiple roles associated with github_user/github_repo
